//--------------------------------------------------------------//
// Shapes
//
// Copyright (c) LWKS Software Ltd.  All Rights Reserved
//--------------------------------------------------------------//
#include "_utils.fx"

int _LwksEffectInfo
<
   string EffectGroup    = "GenericPixelShader";
   string Description    = "Simple 2D Shape";
   string Category       = "Mattes";
   string Notes          = "Rectangles, squares, ellipses and circles";
   bool CanSize          = true;
   bool HasMinOutputSize = true;
> = 0;


//--------------------------------------------------------------//
// Params
//--------------------------------------------------------------//
int SetTechnique
<
   string Description = "Shape";
   string Enum = "Rectangle,Ellipse";
> = 0;

float CentreX
<
   string Description = "Position";
   string Flags = "SpecifiesPointX|DisplayAsPercentage";
   float MinVal = -1.00;
   float MaxVal = 2.00;
> = 0.5;

float CentreY
<
   string Description = "Position";
   string Flags = "SpecifiesPointY|DisplayAsPercentage";
   float MinVal = -1.00;
   float MaxVal = 2.00;
> = 0.5;

float Width
<
   string Description = "Width";
   string Flags = "DisplayAsPercentage";
   float MinVal = 0.00001;
   float MaxVal = 1.00;
> = 0.5;

float Height
<
   string Description = "Height";
   string Flags = "DisplayAsPercentage";
   float MinVal = 0.00001;
   float MaxVal = 1.00;
> = 0.5;

float Softness
<
   string Description = "Softness";
   string Flags = "DisplayAsPercentage";
   float MinVal = 0.00;
   float MaxVal = 0.5;
> = 0.01;

float4 FGColour
<
   string Description = "Colour";
   bool SupportsAlpha = true;
> = { 1.0, 1.0, 1.0, 1.0 };

float  _OutputAspectRatio = 1.0;
float4 _bgExtents;

//--------------------------------------------------------------//
// Inputs
//--------------------------------------------------------------//
DeclareInput( bg, BgSampler );

//--------------------------------------------------------------//
// Code
//--------------------------------------------------------------//
float4 rectangle_main( float2 xy1 : TEXCOORD1 ) : COLOR
{
   float4 bg = ReadPixel( BgSampler, xy1 );

   // Calc exact rectangle bounds

   float l = CentreX - ( Width / _OutputAspectRatio / 2.0 );
   float r = l + Width / _OutputAspectRatio;
   float t = ( 1.0 - CentreY ) - ( Height / 2.0 );
   float b = t + Height;

   float4 ret;

   // Remap the supplied pos so that it's in OUTPUT space, not INPUT space :

   xy1.x = ( xy1.x - _bgExtents[ 0 ] ) / ( _bgExtents[ 2 ] - _bgExtents[ 0 ] );
   xy1.y = ( xy1.y - _bgExtents[ 1 ] ) / ( _bgExtents[ 3 ] - _bgExtents[ 1 ] );

   if ( xy1.x >= l && xy1.x <= r && xy1.y >= t && xy1.y <= b )
   {
      ret = lerp( bg, FGColour, FGColour.a );
   }
   else
   {
      float2 softness = float2( Softness / _OutputAspectRatio, Softness );

      // calc outer bounds

      float l2 = l - softness.x;
      float r2 = r + softness.x;
      float t2 = t - softness.y;
      float b2 = b + softness.y;

      if ( xy1.x < l2 || xy1.x > r2 || xy1.y < t2 || xy1.y > b2 )
      {
         ret = bg;
      }
      else
      {
         float amt = 1.0;

         if ( xy1.x < l )
         {
            if ( xy1.y < t )
            {
               amt = 1.0 - ( length( ( xy1 - float2( l, t ) ) / softness ) );
            }
            else if ( xy1.y > b )
            {
               amt = 1.0 - ( length( ( xy1 - float2( l, b ) ) / softness ) );
            }
            else
               amt = ( xy1.x - l2 ) / softness.x;
         }
         else if ( xy1.x > r )
         {
            if ( xy1.y < t )
            {
               amt = 1.0 - ( length( ( xy1 - float2( r, t ) ) / softness ) );
            }
            else if ( xy1.y > b )
            {
               amt = 1.0 - ( length( ( xy1 - float2( r, b ) ) / softness ) );
            }
            else
               amt = ( r2 - xy1.x ) / softness.x;
         }
         else if ( xy1.y < t )
         {
            amt = ( xy1.y - t2 ) / softness.y;
         }
         else
         {
            amt = ( b2 - xy1.y ) / softness.y;
         }

         ret = lerp( bg, FGColour, max( amt, 0.0 ) * FGColour.a );
      }
   }

   return ret;
}

//--------------------------------------------------------------//
float4 ellipse_main( float2 xy1 : TEXCOORD1 ) : COLOR
{
   float4 bg = ReadPixel( BgSampler, xy1 );

   float a = ( Width/ _OutputAspectRatio ) / 2;
   float b = Height / 2;

   float sa = ( ( Width / _OutputAspectRatio ) / 2 ) + ( Softness / _OutputAspectRatio );
   float sb = ( Height / 2 ) + Softness;

   float2 pos = xy1 - float2( CentreX, 1.0 - CentreY );

   // https://www.mathwarehouse.com/ellipse/equation-of-ellipse.php
   //
   // ( ( x * x ) / ( a * a ) ) + ( ( y * y ) / ( b * b ) ) = 1
   //

   float2 posSq = pos * pos;

   float v1 = ( posSq.x / ( a * a ) )   + ( posSq.y / ( b * b ) );
   float v2 = ( posSq.x / ( sa * sa ) ) + ( posSq.y / ( sb * sb ) );

   float4 ret;

   if ( v1 < 1.0 )
   {
      ret = lerp( bg, FGColour, FGColour.a );
   }
   else if ( v2 > 1.0 )
   {
      ret = bg;
   }
   else
   {
      // http://www.slader.com/discussion/question/what-is-the-equation-of-an-ellipse-in-polar-coordinates/

      float theta = atan2( pos.y, pos.x );

      float cosThetaSq = pow( cos( theta ), 2 );
      float sinThetaSq = pow( sin( theta ), 2 );

      float dLower = ( a * b )   / sqrt( (  b *  b * cosThetaSq ) + (  a *  a * sinThetaSq ) );
      float dUpper = ( sa * sb ) / sqrt( ( sb * sb * cosThetaSq ) + ( sa * sa * sinThetaSq ) );
      float d      = length( pos );

      ret = lerp( bg, FGColour, ( 1.0 - ( ( d - dLower ) / ( dUpper - dLower ) ) ) * FGColour.a );
   }

   return ret;
}

//--------------------------------------------------------------//
// Technique Section for Image Processing Effects
//--------------------------------------------------------------//
technique Rectangle { pass Single_Pass { PixelShader = compile PROFILE rectangle_main(); } }
technique Ellipse   { pass Single_Pass { PixelShader = compile PROFILE ellipse_main(); } }
